@@ -372,62 +372,34 @@ get_device_once(int *master, int *slave, char SlaveName[DEVICELEN], int nomesg,
     int  i;
     char MasterName[DEVICELEN];
 
+#define HEX1(c) \
+	c"0",c"1",c"2",c"3",c"4",c"5",c"6",c"7", \
+	c"8",c"9",c"a",c"b",c"c",c"d",c"e",c"f"
+
 #if defined(__hpux)
     static const char MasterDevice[] = "/dev/ptym/pty%s";
     static const char SlaveDevice[] =  "/dev/pty/tty%s";
     static const char deviceNo[][3] = {
-    "p0","p1","p2","p3","p4","p5","p6","p7",
-    "p8","p9","pa","pb","pc","pd","pe","pf",
-    "q0","q1","q2","q3","q4","q5","q6","q7",
-    "q8","q9","qa","qb","qc","qd","qe","qf",
-    "r0","r1","r2","r3","r4","r5","r6","r7",
-    "r8","r9","ra","rb","rc","rd","re","rf",
-    "s0","s1","s2","s3","s4","s5","s6","s7",
-    "s8","s9","sa","sb","sc","sd","se","sf",
-    "t0","t1","t2","t3","t4","t5","t6","t7",
-    "t8","t9","ta","tb","tc","td","te","tf",
-    "u0","u1","u2","u3","u4","u5","u6","u7",
-    "u8","u9","ua","ub","uc","ud","ue","uf",
-    "v0","v1","v2","v3","v4","v5","v6","v7",
-    "v8","v9","va","vb","vc","vd","ve","vf",
-    "w0","w1","w2","w3","w4","w5","w6","w7",
-    "w8","w9","wa","wb","wc","wd","we","wf",
+	HEX1("p"), HEX1("q"), HEX1("r"), HEX1("s"),
+	HEX1("t"), HEX1("u"), HEX1("v"), HEX1("w"),
     };
 #elif defined(_IBMESA)  /* AIX/ESA */
     static const char MasterDevice[] = "/dev/ptyp%s";
     static const char SlaveDevice[] = "/dev/ttyp%s";
     static const char deviceNo[][3] = {
-    "00","01","02","03","04","05","06","07","08","09","0a","0b","0c","0d","0e","0f",
-    "10","11","12","13","14","15","16","17","18","19","1a","1b","1c","1d","1e","1f",
-    "20","21","22","23","24","25","26","27","28","29","2a","2b","2c","2d","2e","2f",
-    "30","31","32","33","34","35","36","37","38","39","3a","3b","3c","3d","3e","3f",
-    "40","41","42","43","44","45","46","47","48","49","4a","4b","4c","4d","4e","4f",
-    "50","51","52","53","54","55","56","57","58","59","5a","5b","5c","5d","5e","5f",
-    "60","61","62","63","64","65","66","67","68","69","6a","6b","6c","6d","6e","6f",
-    "70","71","72","73","74","75","76","77","78","79","7a","7b","7c","7d","7e","7f",
-    "80","81","82","83","84","85","86","87","88","89","8a","8b","8c","8d","8e","8f",
-    "90","91","92","93","94","95","96","97","98","99","9a","9b","9c","9d","9e","9f",
-    "a0","a1","a2","a3","a4","a5","a6","a7","a8","a9","aa","ab","ac","ad","ae","af",
-    "b0","b1","b2","b3","b4","b5","b6","b7","b8","b9","ba","bb","bc","bd","be","bf",
-    "c0","c1","c2","c3","c4","c5","c6","c7","c8","c9","ca","cb","cc","cd","ce","cf",
-    "d0","d1","d2","d3","d4","d5","d6","d7","d8","d9","da","db","dc","dd","de","df",
-    "e0","e1","e2","e3","e4","e5","e6","e7","e8","e9","ea","eb","ec","ed","ee","ef",
-    "f0","f1","f2","f3","f4","f5","f6","f7","f8","f9","fa","fb","fc","fd","fe","ff",
+	HEX1("0"), HEX1("1"), HEX1("2"), HEX1("3"),
+	HEX1("4"), HEX1("5"), HEX1("6"), HEX1("7"),
+	HEX1("8"), HEX1("9"), HEX1("a"), HEX1("b"),
+	HEX1("c"), HEX1("d"), HEX1("e"), HEX1("f"),
     };
 #else /* 4.2BSD */
     static const char MasterDevice[] = "/dev/pty%s";
     static const char SlaveDevice[] = "/dev/tty%s";
     static const char deviceNo[][3] = {
-    "p0","p1","p2","p3","p4","p5","p6","p7",
-    "p8","p9","pa","pb","pc","pd","pe","pf",
-    "q0","q1","q2","q3","q4","q5","q6","q7",
-    "q8","q9","qa","qb","qc","qd","qe","qf",
-    "r0","r1","r2","r3","r4","r5","r6","r7",
-    "r8","r9","ra","rb","rc","rd","re","rf",
-    "s0","s1","s2","s3","s4","s5","s6","s7",
-    "s8","s9","sa","sb","sc","sd","se","sf",
+	HEX1("p"), HEX1("q"), HEX1("r"), HEX1("s"),
     };
 #endif
+#undef HEX1
     for (i = 0; i < numberof(deviceNo); i++) {
 	const char *const devno = deviceNo[i];
 	snprintf(MasterName, sizeof MasterName, MasterDevice, devno);
