@@ -1,10 +1,10 @@
-#define RUBY_VERSION "2.4.10"
-#define RUBY_RELEASE_DATE "2020-03-31"
-#define RUBY_PATCHLEVEL 364
+#define RUBY_VERSION "2.5.0"
+#define RUBY_RELEASE_DATE RUBY_RELEASE_YEAR_STR"-"RUBY_RELEASE_MONTH_STR"-"RUBY_RELEASE_DAY_STR
+#define RUBY_PATCHLEVEL 0
 
-#define RUBY_RELEASE_YEAR 2020
-#define RUBY_RELEASE_MONTH 3
-#define RUBY_RELEASE_DAY 31
+#define RUBY_RELEASE_YEAR 2017
+#define RUBY_RELEASE_MONTH 12
+#define RUBY_RELEASE_DAY 25
 
 #include "ruby/version.h"
 
